# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import

import calendar
import datetime
import json
import pickle
import re
import pytz
import six
import zlib

from google.protobuf.internal.type_checkers import Int64ValueChecker

from .base.properties import BaseProperty, ContainerBaseProperty
from .datastore.connection import get_connection
from .key import Key


INT_VALUE_CHECKER = Int64ValueChecker()


class BooleanProperty(BaseProperty):
    """A bool property."""
    def from_protobuf(self, pb_value):
        return pb_value.boolean_value

    def to_protobuf(self, value):
        return 'boolean_value', value

    def validate(self, value):
        if not isinstance(value, bool):
            self.error('Value must be a boolean')
        return value


class KeyProperty(BaseProperty):
    """
    The Key for an Entity.

    If this property has no value, an automatic int value will be generated for it and assigned as it's id component.

    This class shouldn't be used directly. An instance of this class is added to the `key` attribute of each
    :class:`~gcloud.base.entity.BaseEntity` instance automatically. You can either let an automatic value be assigned to
    that key or assign a value manually as follows:

        >>> e = Entity(key=1)  # e has a KeyProperty at .key with an id component of 1
        >>> e = Entity(key='1')  # e has a KeyProperty at .key with an name component of '1'
        >>> e = Entity()  # e will have a KeyProperty at .key with a generated id component when .save() is called
        >>> e.key = 1  # e now has a KeyProperty at .key with an id component of 1

    In Datastore, all keys must have a kind. The kind is automatically set the the ``__class__`` attribute of the owning
    :class:`~gcloud.base.entity.BaseEntity` subclass. The value for this property is a dict containing the kind and the
    id (int) or name (str in Python 3, unicode in Python 2). For example:

        >>> e = Entity(key=1)
        >>> e.key
        {'kind': 'Entity', 'id': 1}
        >>> e = Entity(key='1')
        >>> e.key
        {'kind': 'Entity', 'name': 1}
        >>> e = Entity()
        >>> e.key
        {'kind': 'Entity'}

    As you can see, when a key has an auto generated id, the id component will remain empty until the entity is saved.

    Even though the value of a key is a dict, you must pass either an int or str (unicode in Python 2) to set its value:

        >>> e = Entity()
        >>> e.key
        {'kind': 'Entity'}
        >>> e.key = 1
        >>> e.key
        {'kind': 'Entity', 'id': 1}
    """
    def __init__(self, name=None, db_name=None):
        super(KeyProperty, self).__init__(name=name, db_name=db_name)

    def __get__(self, instance, owner):
        if not instance:
            return self
        return instance._data.get(self.name)

    def __set__(self, instance, value):
        kind = instance._meta.kind
        parent = None

        # Is this a (parent, <value>) tuple?
        if isinstance(value, tuple):
            parent = value[0]
            if len(value) == 2:
                value = value[1]  # Actual key value as 2nd element
            else:
                value = None  # Partial key

        if isinstance(value, Key):
            instance._data[self.name] = value
        else:
            instance._data[self.name] = Key(kind, parent=parent, value=value)

    def to_protobuf(self, value):
        from .datastore import datastore_v1_pb2 as datastore_pb
        key = datastore_pb.Key()

        dataset_id = get_connection().dataset
        if not dataset_id:
            raise EnvironmentError("Couldn't determine the dataset ID. Have you called connect?")
        key.partition_id.dataset_id = dataset_id
        key.partition_id.namespace = get_connection().namespace

        for item in value.path:
            element = key.path_element.add()
            element.kind = item.kind
            if item.id:
                element.id = item.id
            elif item.name:
                element.name = item.name

        return key

    def from_protobuf(self, pb_value):
        """
        Factory method for creating a key based on a protobuf.

        The protobuf should be one returned from the Datastore protobuf API.

        :type pb_value: :class:`gcloudoem.datastore.datastore_v1_pb2.Key`
        :param pb_value: The Protobuf representing the key.

        :rtype: :class:`gcloudoem.key.Key`
        :returns: a new `Key` instance
        """
        last_key = None
        for element in pb_value.path_element:
            if element.HasField('id'):
                key = Key(element.kind, parent=last_key, value=element.id)
            elif element.HasField('name'):
                key = Key(element.kind, parent=last_key, value=element.name)
            else:
                key = Key(element.kind, parent=last_key)
            last_key = key

        return last_key


class ReferenceProperty(BaseProperty):
    """
    A reference to another entity.

    The value of this property is an Entity instance. It is required by default.
    """
    def __init__(self, entity_cls, name=None, db_name=None, required=True, verbose_name=None):
        """
        Initialise a new reference property.

        :param entity_cls: The class of the entity this property is referencing.
        """
        super(ReferenceProperty, self).__init__(name=name, db_name=db_name, required=required)
        self.entity_cls = entity_cls

    def __get__(self, instance, owner):
        """Fetch this field from the QuerySet if not already stored on the class."""
        if instance is None:
            return self
        try:
            value = instance._data[self.name]
            if isinstance(value, Key):  # We need to fetch the entity and set it on the owning entity
                value = self.entity_cls.objects.get(pk=value.name_or_id)
                setattr(instance, self.name, value)
        except KeyError:  # Empty
            pass
        return instance._data.get(self.name)

    def __set__(self, instance, value):
        if not isinstance(value, (Key, self.entity_cls)) and value is not None:
            raise TypeError('The value of a ReferenceProperty must be an Entity or Key')
        instance._data[self.name] = value

    def from_protobuf(self, pb_value):
        data = pb_value.blob_value
        return pickle.loads(data)

    def to_protobuf(self, value):
        if isinstance(value, self.entity_cls):
            value = value.key
        return "blob_value", pickle.dumps(value, protocol=2)  # Py2 compatible

    def validate(self, value):
        if not isinstance(value, (self.entity_cls, Key)):
            self.error("A ReferenceProperty only accepts an %s instance or a Key as its value." %
                       self.entity_cls.__name__)
        if (isinstance(value, self.entity_cls) and value.key.is_partial) or \
                (isinstance(value, Key) and value.is_partial):
            self.error("You can only reference an Entity once it has been saved to Datastore.")


class IntegerProperty(BaseProperty):
    """An int property."""
    def from_protobuf(self, pb_value):
        return pb_value.integer_value

    def to_protobuf(self, value):
        INT_VALUE_CHECKER.CheckValue(value)
        return 'integer_value', int(value)

    def validate(self, value):
        if not isinstance(value, six.integer_types):
            self.error('Value must be an int (or long in Python 2).')
        return int(value)


class FloatProperty(BaseProperty):
    """A float property."""
    def from_protobuf(self, pb_value):
        return pb_value.double_value

    def to_protobuf(self, value):
        return 'double_value', value

    def validate(self, value):
        if not isinstance(value, six.integer_types + (float,)):
            self.error('Value must be an int or float (or long in Python 2).')
        return float(value)


class BlobProperty(BaseProperty):
    """Store data as bytes. Supports compression."""
    def __init__(self, compressed=False, **kwargs):
        """
        Initialise this property. Has an option to compress using zlib that defaults to False. **Note** that this
        property can't be compressed and indexed!

        :param bool compressed: should this property store its value compressed? Defaults to False.
        """
        super(BlobProperty, self).__init__(**kwargs)

        self._compressed = compressed

    def from_protobuf(self, pb_value):
        value = pb_value.blob_value
        if self._compressed:
            return zlib.decompress(value)
        return value

    def to_protobuf(self, value):
        if self._compressed:
            value = zlib.compress(value)
        return 'blob_value', value

    def validate(self, value):
        if not isinstance(value, six.binary_type):
            self.error("Value must be bytes (or str in Python 2).")
        return value


class TextProperty(BlobProperty):
    """Store data as unicode."""
    def __init__(self, max_length=None, **kwargs):
        super(TextProperty, self).__init__(**kwargs)
        self._max_length = max_length

    @property
    def max_length(self):
        return self._max_length

    def from_protobuf(self, pb_value):
        value = pb_value.string_value
        if isinstance(value, six.binary_type):
            return value.encode('utf-8')
        return value

    def to_protobuf(self, value):
        if isinstance(value, six.binary_type):
            value = value.decode('utf-8')
        return 'string_value', value

    def validate(self, value):
        if not isinstance(value, six.string_types):
            self.error('Value must be str (unicode in Python 2)')
        if self.max_length and len(value) > self.max_length:
            self.error('Value exceeds maximum length of %d' % self.max_length)
        return value


class EmailProperty(TextProperty):
    message = 'Enter a valid email address.'
    user_regex = re.compile(
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*$"  # dot-atom
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"$)',  # quoted-string
        re.IGNORECASE
    )
    domain_regex = re.compile(
        # max length for domain name labels is 63 characters per RFC 1034
        r'((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+)(?:[A-Z0-9-]{2,63}(?<!-))$',
        re.IGNORECASE
    )
    literal_regex = re.compile(
        # literal form, ipv4 or ipv6 address (SMTP 4.1.3)
        r'\[([A-f0-9:\.]+)\]$',
        re.IGNORECASE
    )
    ipv4 = re.compile(r'^(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}$')
    domain_whitelist = ['localhost']

    def validate(self, value):
        if not value or '@' not in value:
            self.error(self.message)

        user_part, domain_part = value.rsplit('@', 1)

        if not self.user_regex.match(user_part):
            self.error(self.message)

        if domain_part not in self.domain_whitelist and not self.validate_domain_part(domain_part):
            # Try for possible IDN domain-part
            try:
                domain_part = domain_part.encode('idna').decode('ascii')
                if self.validate_domain_part(domain_part):
                    return
            except UnicodeError:
                pass
            self.error(self.message)

    def validate_domain_part(self, domain_part):
        from .exceptions import ValidationError

        if self.domain_regex.match(domain_part):
            return True

        literal_match = self.literal_regex.match(domain_part)
        if literal_match:
            ip_address = literal_match.group(1)
            if not self.is_valid_ipv4(ip_address):
                if not self.is_valid_ipv6(ip_address):
                    return False
        return True

    def is_valid_ipv4(self, ip_str):
        return bool(self.ipv4.search(ip_str))

    def is_valid_ipv6(self, ip_str):
        # We need to have at least one ':'.
        if ':' not in ip_str:
            return False

        # We can only have one '::' shortener.
        if ip_str.count('::') > 1:
            return False

        # '::' should be encompassed by start, digits or end.
        if ':::' in ip_str:
            return False

        # A single colon can neither start nor end an address.
        if ((ip_str.startswith(':') and not ip_str.startswith('::')) or
                (ip_str.endswith(':') and not ip_str.endswith('::'))):
            return False

        # We can never have more than 7 ':' (1::2:3:4:5:6:7:8 is invalid)
        if ip_str.count(':') > 7:
            return False

        # If we have no concatenation, we need to have 8 fields with 7 ':'.
        if '::' not in ip_str and ip_str.count(':') != 7:
            # We might have an IPv4 mapped address.
            if ip_str.count('.') != 3:
                return False

        ip_str = self._explode_shorthand_ip_string(ip_str)

        # Now that we have that all squared away, let's check that each of the
        # hextets are between 0x0 and 0xFFFF.
        for hextet in ip_str.split(':'):
            if hextet.count('.') == 3:
                # If we have an IPv4 mapped address, the IPv4 portion has to
                # be at the end of the IPv6 portion.
                if not ip_str.split(':')[-1] == hextet or not self.is_valid_ipv4(hextet):
                    return False
            else:
                try:
                    # a value error here means that we got a bad hextet,
                    # something like 0xzzzz
                    if int(hextet, 16) < 0x0 or int(hextet, 16) > 0xFFFF:
                        return False
                except ValueError:
                    return False

    def _explode_shorthand_ip_string(self, ip_str):
        """
        Expand a shortened IPv6 address.
        Args:
            ip_str: A string, the IPv6 address.
        Returns:
            A string, the expanded IPv6 address.
        """
        if not self._is_shorthand_ip(ip_str):
            # We've already got a longhand ip_str.
            return ip_str

        new_ip = []
        hextet = ip_str.split('::')

        # If there is a ::, we need to expand it with zeroes
        # to get to 8 hextets - unless there is a dot in the last hextet,
        # meaning we're doing v4-mapping
        if '.' in ip_str.split(':')[-1]:
            fill_to = 7
        else:
            fill_to = 8

        if len(hextet) > 1:
            sep = len(hextet[0].split(':')) + len(hextet[1].split(':'))
            new_ip = hextet[0].split(':')

            for __ in range(fill_to - sep):
                new_ip.append('0000')
            new_ip += hextet[1].split(':')

        else:
            new_ip = ip_str.split(':')

        # Now need to make sure every hextet is 4 lower case characters.
        # If a hextet is < 4 characters, we've got missing leading 0's.
        ret_ip = []
        for hextet in new_ip:
            ret_ip.append(('0' * (4 - len(hextet)) + hextet).lower())
        return ':'.join(ret_ip)

    def _is_shorthand_ip(ip_str):
        """
        Determine if the address is shortened.

        Args:
            ip_str: A string, the IPv6 address.
        Returns:
            A boolean, True if the address is shortened.
        """
        if ip_str.count('::') == 1:
            return True
        if any(len(x) < 4 for x in ip_str.split(':')):
            return True
        return False


class PickleProperty(BlobProperty):
    """Store data as pickle. Takes care of (un)pickling."""
    def to_protobuf(self, value):
        return super(PickleProperty, self).to_protobuf(pickle.dumps(value, protocol=2))  # Py2 compatible

    def from_protobuf(self, pb_value):
        return pickle.loads(super(PickleProperty, self).from_protobuf(pb_value))


class DictProperty(PickleProperty):
    """
    Store a python dict.

    Dict is pickled for storage.
    """
    def validate(self, value):
        if not isinstance(value, dict):
            self.error("Value for a DictProperty must be a dict instance. Got %s instead." % type(value))


class JsonProperty(BlobProperty):
    """Store data as JSON. Takes care of conversion to/from JSON."""
    def __init__(self, name=None, schema=None, **kwargs):
        super(JsonProperty, self).__init__(name, **kwargs)
        self._schema = schema

    def to_datastore_value(self, value):
        return super(JsonProperty, self)._to_base_type(json.dumps(value))

    def to_python_value(self, value):
        return json.loads(super(JsonProperty, self)._from_base_type(value))


class DateTimeProperty(BaseProperty):
    """Store data as a timestamp represented as datetime.datetime."""
    def __init__(self, name=None, auto_now_add=False, auto_now=False, **kwargs):
        assert not ((auto_now_add or auto_now) and kwargs.get("repeated", False))
        super(DateTimeProperty, self).__init__(name, **kwargs)
        self._auto_now_add = auto_now_add
        self._auto_now = auto_now

    def from_protobuf(self, pb_value):
        microseconds = pb_value.timestamp_microseconds_value
        naive = (datetime.datetime.utcfromtimestamp(0) + datetime.timedelta(microseconds=microseconds))
        return naive.replace(tzinfo=pytz.utc)

    def to_protobuf(self, value):
        name = 'timestamp_microseconds_value'
        # If the datetime is naive (no timezone), consider that it was
        # intended to be UTC and replace the tzinfo to that effect.
        if not value.tzinfo:
            value = value.replace(tzinfo=pytz.utc)
        # Regardless of what timezone is on the value, convert it to UTC.
        value = value.astimezone(pytz.utc)
        # Convert the datetime to a microsecond timestamp.
        value = int(calendar.timegm(value.timetuple()) * 1e6) + value.microsecond
        return name, value

    def validate(self, value):
        if not isinstance(value, datetime.datetime):
            self.error('Value must be a datetime.datetime')
        return value

    def _now(self):
        return datetime.datetime.utcnow()


class DateProperty(DateTimeProperty):
    """Store data as a date and represented as datetime.date."""
    def validate(self, value):
        if not isinstance(value, datetime.date):
            self.error('Value must be a datetime.date')
        return value

    def from_protobuf(self, pb_value):
        microseconds = pb_value.timestamp_microseconds_value
        naive = (datetime.datetime.utcfromtimestamp(0) + datetime.timedelta(microseconds=microseconds))
        return naive.replace(tzinfo=pytz.utc).date()

    def to_protobuf(self, value):
        name = 'timestamp_microseconds_value'
        # If the datetime is naive (no timezone), consider that it was
        # intended to be UTC and replace the tzinfo to that effect.
        if not value.tzinfo:
            value = value.replace(tzinfo=pytz.utc)
        # Regardless of what timezone is on the value, convert it to UTC.
        value = value.astimezone(pytz.utc)
        # Convert the datetime to a microsecond timestamp.
        value = int(calendar.timegm(value.timetuple()) * 1e6) + value.microsecond
        return name, value

    def _now(self):
        return datetime.datetime.utcnow().date()


class TimeProperty(DateTimeProperty):
    """Store data as time represented using datetime.time."""
    def _validate(self, value):
        if not isinstance(value, datetime.time):
            self.error("Value must be a datetime.time")
        return value

    def from_protobuf(self, pb_value):
        microseconds = pb_value.timestamp_microseconds_value
        naive = (datetime.datetime.utcfromtimestamp(0) + datetime.timedelta(microseconds=microseconds))
        return naive.replace(tzinfo=pytz.utc).time()

    def to_protobuf(self, value):
        name = 'timestamp_microseconds_value'
        # If the datetime is naive (no timezone), consider that it was
        # intended to be UTC and replace the tzinfo to that effect.
        if not value.tzinfo:
            value = value.replace(tzinfo=pytz.utc)
        # Regardless of what timezone is on the value, convert it to UTC.
        value = value.astimezone(pytz.utc)
        # Convert the datetime to a microsecond timestamp.
        value = int(calendar.timegm(value.timetuple()) * 1e6) + value.microsecond
        return name, value


class ListProperty(ContainerBaseProperty):
    """
    A property that supports a list of properties.

    Only supports one type of property at a time.

    .. note::
        Required means it cannot be empty - as the default for ListProperty is []
    """
    def __init__(self, property, **kwargs):
        """
        :param :class:`~gcloudoem.base.properties.BaseProperty` property: The property class used as the value of the
            ``list`` items. Can't be a :class:`~gcloudoem.properties.KeyProperty`
        """
        if not isinstance(property, BaseProperty):
            raise TypeError('property must be a BaseProperty Instance')
        if isinstance(property, (KeyProperty, ListProperty)):
            raise TypeError('property cannot be a KeyProperty or ListProperty')
        self.property = property
        kwargs.pop('default', lambda: [])
        super(ListProperty, self).__init__(**kwargs)

    def from_protobuf(self, pb_value):
        return [self.property.from_protobuf(v) for v in pb_value.list_value]

    def to_protobuf(self, value):
        return 'list_value', [self.property.to_protobuf(v) for v in value]

    def validate(self, value):
        if not isinstance(value, (list, tuple)):
            self.error('Value must be a list or tuple')
        for item in value:
            self.property.validate(item)



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
"""
:class:`.Entity`s are the building blocks of the OEM layer, providing object-orientated interaction with datastore
entities.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import with_metaclass

from .base.entity import BaseEntity
from .base.metaclasses import EntityMeta
from .datastore.transaction import Transaction


class Entity(with_metaclass(EntityMeta, BaseEntity)):
    """
    A :class:`Entity` is just a mapping of an object to a datastore entity. ``cls.__name__`` is used as the kind for
    the datastore entity. A :class:`Entity` has 1 or more properties as its attributes which are persisted to the
    datastore as  properties of a datastore entity. See :mod:`gcloudoem.properties` for a full list of available
    properties. For example:

        from gcloudoem import entity, properties

        class Person(entity.Entity):
            name = properties.TextProperty(indexed=False)
            dob = properties.DateProperty()

    The id/name for the key of the underlying entity will be automatically inferred from the Entity instance using the
    following strategy:

    * A :class:'~gcloudoem.properties.KeyProperty' stored on the ``key`` attribute. Otherwise;
    * A ``key`` attribute of type :class:`~gcloudoem.properties.KeyProperty` will be injected onto the instance.

    If more then one :class:'~gcloudoem.properties.KeyProperty' is present on the Entity, an AttributeError will be
    thrown.

    Access to the entity's key is via :attr:`.key`, and the id can be fetched using ``.key.id_or_name``.

    To save/update an model, call :func:`~.save` on it. To fetch a model by id, call :func:`~.get_by_id`. To fetch
    multiple model instances at once, use :func:`~.filter`. To delete a model instance from datastore, call :
    func:`~.delete`.

    This class shouldn't be used directly. Instead, it is intended to be extended by concrete model implementations.
    """
    def __init__(self, **kwargs):
        super(Entity, self).__init__(**kwargs)

    def save(self, force_insert=False, validate=True, clean=True, **kwargs):
        """
        Save the :class:`Entity` to the database. If the entity already exists, it will be updated,
        otherwise it will be created.

        :param force_insert: only try to create a new document, don't allow updates of existing documents. Defaults to
            False.
        :param validate: validates the document; set to ``False`` to skip.
        :param clean: call the document clean method, requires `validate` to be True.

        """
        if validate:
            self.validate(clean=clean)

        with Transaction(Transaction.SNAPSHOT) as txn:
            if force_insert:
                txn.create(self)
            else:
                txn.put(self)

    def delete(self):
        """Delete this entity from Datastore."""
        with Transaction(Transaction.SNAPSHOT) as txn:
            txn.delete(self)

    @classmethod
    def from_protobuf(cls, pb):
        key = None
        if pb.HasField('key'):
            key = cls._properties['key'].from_protobuf(pb.key)

        entity_props = {
            'key': key
        }

        for property_pb in pb.property:
            if not hasattr(cls, property_pb.name):
                raise ValueError("Entity %s doesn't have a property by the name %s" %
                                 (cls._meta.kind, property_pb.name))
            value = cls._properties[property_pb.name].from_protobuf(property_pb.value)
            entity_props[property_pb.name] = value

        instance = cls(**entity_props)
        return instance

    def __repr__(self):
        key = self._data['key']
        return "<%s (%s) %s)>" % (
            self.__class__.__name__,
            key.name_or_id,
            super(Entity, self).__repr__()
        )



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
import re


VERSION_PICKLE_KEY = '_gcloudoem_version'
re_camel_case = re.compile(r'(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))')


def camel_case_to_spaces(value):
    """Splits CamelCase and converts to lower case. Also strips leading and trailing whitespace."""
    return re_camel_case.sub(r' \1', value).strip().lower()



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGED BY Kapiche Ltd.
# Copyright 2015 Kapiche Ltd. All rights reserved.
# Based on work by the good folk responsible for gcloud-python. Thanks folks!
# Author: Ryan Stuart<ryan@kapiche.com>
#
"""A simple wrapper around the OAuth2 credentials library."""
from __future__ import absolute_import, division, print_function, unicode_literals

import base64
import calendar
import datetime

from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from oauth2client import client
from oauth2client.client import _get_application_default_credential_from_file
from oauth2client import crypt
from oauth2client import service_account
import pytz
import six
from six.moves.urllib.parse import urlencode


def get_credentials():
    """
    Gets credentials implicitly from the current environment.

    .. note::
      You should not need to use this function directly. Instead, use the helper method
      :func:`gcloudoem.datastore.__init__.get_connection` which uses this method under the hood.

    Checks environment in order of precedence:

    * Google App Engine (production and testing)
    * Environment variable GOOGLE_APPLICATION_CREDENTIALS pointing to a file with stored credentials information.
    * Stored "well known" file associated with ``gcloud`` command line tool.
    * Google Compute Engine production environment.

    The file referred to in GOOGLE_APPLICATION_CREDENTIALS is expected to contain information about credentials that are
    ready to use. This means either service account information or user account information with a ready-to-use refresh
    token::

      {                                       {
          'type': 'authorized_user',              'type': 'service_account',
          'client_id': '...',                     'client_id': '...',
          'client_secret': '...',       OR        'client_email': '...',
          'refresh_token': '...,                  'private_key_id': '...',
      }                                           'private_key': '...',
                                              }

    The second of these is simply a JSON key downloaded from the Google APIs console. The first is a close cousin of the
    "client secrets" JSON file used by ``oauth2client.clientsecrets`` but differs in formatting.

    :rtype: :class:`oauth2client.client.GoogleCredentials`,
            :class:`oauth2client.appengine.AppAssertionCredentials`,
            :class:`oauth2client.gce.AppAssertionCredentials`,
            :class:`oauth2client.service_account._ServiceAccountCredentials`
    :returns: A new credentials instance corresponding to the implicit environment.
    """
    return client.GoogleCredentials.get_application_default()


def get_for_service_account_json(json_credentials_path, scope=None):
    """
    Gets the credentials for a service account with JSON key.

    :param str json_credentials_path: The path to a private key file (this file was given to you when you created the
        service account). This file must contain a JSON object with a private key and other credentials information
        (downloaded from the Google APIs console).

    :type scope: string or tuple of string
    :param scope: The scope against which to authenticate. (Different services require different scopes, check the
        documentation for which scope is required for the different levels of access to any particular API.)

    :rtype: :class:`oauth2client.client.GoogleCredentials`,
        :class:`oauth2client.service_account._ServiceAccountCredentials`
    :returns: New service account or Google (for a user JSON key file) credentials object.
    """
    credentials = _get_application_default_credential_from_file(json_credentials_path)
    if scope is not None:
        credentials = credentials.create_scoped(scope)
    return credentials


def get_for_service_account_p12(client_email, private_key_path, scope=None):
    """Gets the credentials for a service account with PKCS12 / p12 key.

    .. note::
      This method is not used by default, instead :func:`get_credentials` is used. This method is intended to be used
      when the environments is known explicitly and detecting the environment implicitly would be superfluous.

    :param str client_email: The e-mail attached to the service account.

    :param str private_key_path: The path to a private key file (this file was given to you when you created the service
        account). This file must be in P12 format.

    :type scope: string or tuple of string
    :param scope: The scope against which to authenticate. (Different services require different scopes, check the
        documentation for which scope is required for the different levels of access to any particular API.)

    :rtype: :class:`oauth2client.client.SignedJwtAssertionCredentials`
    :returns: A new ``SignedJwtAssertionCredentials`` instance with the needed service account settings.
    """
    return client.SignedJwtAssertionCredentials(
        service_account_name=client_email,
        private_key=open(private_key_path, 'rb').read(),
        scope=scope
    )


def _get_pem_key(credentials):
    """Gets RSA key for a PEM payload from a credentials object.

    :type credentials: :class:`client.SignedJwtAssertionCredentials`,
                       :class:`service_account._ServiceAccountCredentials`
    :param credentials: The credentials used to create an RSA key for signing text.

    :rtype: :class:`Crypto.PublicKey.RSA._RSAobj`
    :returns: An RSA object used to sign text.
    :raises: `TypeError` if `credentials` is the wrong type.
    """
    if isinstance(credentials, client.SignedJwtAssertionCredentials):
        # Take our PKCS12 (.p12) key and make it into a RSA key we can use.
        pem_text = crypt.pkcs12_key_as_pem(credentials.private_key, credentials.private_key_password)
    elif isinstance(credentials, service_account._ServiceAccountCredentials):
        pem_text = credentials._private_key_pkcs8_text
    else:
        raise TypeError((credentials, 'not a valid service account credentials type'))

    return RSA.importKey(pem_text)


def _get_signed_query_params(credentials, expiration, signature_string):
    """Gets query parameters for creating a signed URL.

    :type credentials: :class:`client.SignedJwtAssertionCredentials`,
                       :class:`service_account._ServiceAccountCredentials`
    :param credentials: The credentials used to create an RSA key for signing text.

    :type expiration: int or long
    :param expiration: When the signed URL should expire.

    :param str signature_string: The string to be signed by the credentials.

    :rtype: dict
    :returns: Query parameters matching the signing credentials with a signed payload.
    """
    pem_key = _get_pem_key(credentials)
    # Sign the string with the RSA key.
    signer = PKCS1_v1_5.new(pem_key)
    signature_hash = SHA256.new(signature_string)
    signature_bytes = signer.sign(signature_hash)
    signature = base64.b64encode(signature_bytes)

    if isinstance(credentials, client.SignedJwtAssertionCredentials):
        service_account_name = credentials.service_account_name
    elif isinstance(credentials, service_account._ServiceAccountCredentials):
        service_account_name = credentials._service_account_email
    # We know one of the above must occur since `_get_pem_key` fails if not.
    return {
        'GoogleAccessId': service_account_name,
        'Expires': str(expiration),
        'Signature': signature,
    }


def _utcnow():  # pragma: NO COVER testing replaces
    """Returns current time as UTC datetime.

    NOTE: on the module namespace so tests can replace it.
    """
    return datetime.datetime.utcnow()


def _get_expiration_seconds(expiration):
    """
    Convert 'expiration' to a number of seconds in the future.

    :type expiration: int, long, datetime.datetime, datetime.timedelta
    :param expiration: When the signed URL should expire.

    :rtype: int
    :returns: a timestamp as an absolute number of seconds.
    """
    # If it's a timedelta, add it to `now` in UTC.
    if isinstance(expiration, datetime.timedelta):
        now = _utcnow().replace(tzinfo=pytz.utc)
        expiration = now + expiration

    # If it's a datetime, convert to a timestamp.
    if isinstance(expiration, datetime.datetime):
        # Make sure the timezone on the value is UTC
        # (either by converting or replacing the value).
        if expiration.tzinfo:
            expiration = expiration.astimezone(pytz.utc)
        else:
            expiration = expiration.replace(tzinfo=pytz.utc)

        # Turn the datetime into a timestamp (seconds, not microseconds).
        expiration = int(calendar.timegm(expiration.timetuple()))

    if not isinstance(expiration, six.integer_types):
        raise TypeError('Expected an integer timestamp, datetime, or ' 'timedelta. Got %s' % type(expiration))
    return expiration


def generate_signed_url(credentials, resource, expiration, api_access_endpoint='', method='GET', content_md5=None,
                        content_type=None):
    """Generate signed URL to provide query-string auth'n to a resource.

    :type credentials: :class:`oauth2client.appengine.AppAssertionCredentials`
    :param credentials: Credentials object with an associated private key to sign text.

    :param str resource: A pointer to a specific resource (typically, ``/bucket-name/path/to/blob.txt``).

    :type expiration: int, long, datetime.datetime, datetime.timedelta
    :param expiration: When the signed URL should expire.

    :param str api_access_endpoint: Optional URI base. Defaults to empty string.

    :param str method: The HTTP verb that will be used when requesting the URL.

    :param str content_md5: The MD5 hash of the object referenced by ``resource``.

    :param str content_type: The content type of the object referenced by ``resource``.

    :rtype: string
    :returns: A signed URL you can use to access the resource until expiration.
    """
    expiration = _get_expiration_seconds(expiration)

    # Generate the string to sign.
    signature_string = '\n'.join([
        method,
        content_md5 or '',
        content_type or '',
        str(expiration),
        resource]
    )

    # Set the right query parameters.
    query_params = _get_signed_query_params(credentials, expiration, signature_string)

    # Return the built URL.
    return '{endpoint}{resource}?{querystring}'.format(
        endpoint=api_access_endpoint, resource=resource, querystring=urlencode(query_params)
    )



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import
import sys

from ..exceptions import DoesNotExist, MultipleObjectsReturned
from ..options import Options
from ..queryset.manager import QuerySetManager
from .properties import BaseProperty


ENTITY_MODULE_NAMES = (
    'models',  # Django style
    'entities',
)


def subclass_exception(name, parents, module, attached_to=None):
    """
    Create exception subclass. Used by EntityMeta below.

    If 'attached_to' is supplied, the exception will be created in a way that allows it to be pickled, assuming the
    returned exception class will be added as an attribute to the 'attached_to' class.
    """
    class_dict = {'__module__': module}
    if attached_to is not None:
        def __reduce__(self):
            # Exceptions are special - they've got state that isn't
            # in self.__dict__. We assume it is all in self.args.
            return (unpickle_inner_exception, (attached_to, name), self.args)

        def __setstate__(self, args):
            self.args = args

        class_dict['__reduce__'] = __reduce__
        class_dict['__setstate__'] = __setstate__

    return type(name, parents, class_dict)


def unpickle_inner_exception(klass, exception_name):
    # Get the exception class from the class it is attached to:
    exception = getattr(klass, exception_name)
    return exception.__new__(exception)


class EntityMeta(type):
    """
    Metaclass for :class:`~gcloud.entity.Entity` classes.

    Sets the name of :class:`~gcloudoem.base.base.BasePropery` class attributes and injects the
    :class:`~gcloud.properties.KeyProperty` property at ``key`` if required.
    """
    def __new__(cls, name, bases, attrs):
        flattened_bases = cls._get_bases(bases)

        # Store the properties for this entity
        from ..properties import KeyProperty
        properties = {}
        # Merge all properties from subclasses
        for base in flattened_bases[::-1]:
            if hasattr(base, '_properties'):
                properties.update(base._properties)
            # Standard object mixin - merge in any Properties
            if not hasattr(base, '_meta'):
                for attr_name, attr_value in base.__dict__.items():
                    if not isinstance(attr_value, BaseProperty):
                        continue
                    attr_value.name = attr_name
                    if not attr_value.db_name:
                        attr_value.db_name = attr_name
                    properties[attr_name] = attr_value
        # Now merge the properties from the defined Entity
        for prop_name, value in attrs.items():
            if isinstance(value, BaseProperty):
                if prop_name == 'key' and not isinstance(value, KeyProperty):
                    raise AttributeError("Attribute of 'key' isn't allowed unless it is a KeyProperty.")
                elif isinstance(value, KeyProperty) and not prop_name == 'key':
                    raise AttributeError("Only attr 'key' can be a KeyProperty.")
                value.name = prop_name
                if not value.db_name:
                    value.db_name = prop_name
                properties[prop_name] = value
        if 'key' not in attrs:  # Ensure there is a key
            value = KeyProperty(name='key', db_name='__key__')
            attrs['key'] = properties['key'] = value

        attrs['_properties'] = properties

        # Create the class
        module = attrs.pop('__module__')
        new_cls = super(EntityMeta, cls).__new__(cls, name, bases, attrs)

        # Provide a default queryset unless exists or one has been set
        if 'objects' not in dir(new_cls):
            new_cls.objects = QuerySetManager()

        # Build the Meta for this entity
        attr_meta = attrs.pop('Meta', None)
        if not attr_meta:
            meta = getattr(new_cls, 'Meta', None)
        else:
            meta = attr_meta

        # Try to detect the app_label
        if getattr(meta, 'app_label', None) is None:
            entity_module = sys.modules[new_cls.__module__]
            package_components = entity_module.__name__.split('.')
            package_components.reverse()  # find the last occurrence of 'models' or 'entites'
            app_label_index = -1
            for name in ENTITY_MODULE_NAMES:
                try:
                    app_label_index = package_components.index(name)
                    break
                except ValueError:
                    pass
            if app_label_index > -1:
                setattr(meta, "app_label", package_components[app_label_index + 1])

        _meta = Options(meta)
        _meta.contribute_to_class(new_cls, '_meta')

        # Add some exceptions to the Entity.
        setattr(
            new_cls,
            'DoesNotExist',
            subclass_exception(str('DoesNotExist'), (DoesNotExist,), module, attached_to=new_cls)
        )
        setattr(
            new_cls,
            'MultipleObjectsReturned',
            subclass_exception(str('MultipleObjectsReturned'), (MultipleObjectsReturned,), module, attached_to=new_cls)
        )

        return new_cls

    @classmethod
    def _get_bases(cls, bases):
        if isinstance(bases, BasesTuple):
            return bases
        seen = []
        bases = cls.__get_bases(bases)
        unique_bases = (b for b in bases if not (b in seen or seen.append(b)))
        return BasesTuple(unique_bases)

    @classmethod
    def __get_bases(cls, bases):
        for base in bases:
            if base is object:
                continue
            yield base
            for child_base in cls.__get_bases(base.__bases__):
                yield child_base


class BasesTuple(tuple):
    """Special class to handle introspection of bases tuple in __new__"""
    pass



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import, division, print_function, unicode_literals

from .connection import ConnectionError, Connection, register_connection, DEFAULT_NAMESPACE
from .environment import determine_default_dataset_id
from gcloudoem.datastore import credentials


SCOPE = ('https://www.googleapis.com/auth/datastore', 'https://www.googleapis.com/auth/userinfo.email')
"""The scopes required for authenticating as a Cloud Datastore consumer."""


def connect(dataset_id=None, namespace=DEFAULT_NAMESPACE):
    """
    Connect to Datastore. If no dataset is given, we attempt to determine it given the environment.

    :param namespace: The namespace to use. A namespace is used for multi-tenancy on datastore. It's useful for
        separating dev data from production data for example.
    :param str dataset_id: Optional. The dataset ID to use as default.
    """
    if dataset_id is None:
        dataset_id = determine_default_dataset_id()
        if dataset_id is None:
            raise ConnectionError("Couldn't determine the dataset id from the environment")
    implicit_credentials = credentials.get_credentials()
    scoped_credentials = implicit_credentials.create_scoped(SCOPE)
    register_connection(dataset_id, namespace, scoped_credentials)



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGED BY Kapiche Ltd.
# Copyright 2015 Kapiche Ltd. All rights reserved.
# Based on work by the good folk responsible for gcloud-python. Thanks folks!
# Author: Ryan Stuart<ryan@kapiche.com>
#
"""Create / interact with gcloud datastore queries."""
from __future__ import absolute_import, division, print_function

import base64

import six

from . import datastore_v1_pb2 as datastore_pb, utils
from .connection import get_connection
from ..exceptions import InvalidQueryError
from ..key import Key
from .transaction import Transaction


class Query(object):
    """
    A Google Datastore Query.

    This class serves as an abstraction for creating a query over data stored in Datastore.
    """

    OPERATORS = {
        '<=': datastore_pb.PropertyFilter.LESS_THAN_OR_EQUAL,
        'lte': datastore_pb.PropertyFilter.LESS_THAN_OR_EQUAL,
        '>=': datastore_pb.PropertyFilter.GREATER_THAN_OR_EQUAL,
        'gte': datastore_pb.PropertyFilter.GREATER_THAN_OR_EQUAL,
        '<': datastore_pb.PropertyFilter.LESS_THAN,
        'lt': datastore_pb.PropertyFilter.LESS_THAN,
        '>': datastore_pb.PropertyFilter.GREATER_THAN,
        'gt': datastore_pb.PropertyFilter.GREATER_THAN,
        '=': datastore_pb.PropertyFilter.EQUAL,
        'eq': datastore_pb.PropertyFilter.EQUAL,
    }
    """Mapping of operator strs and their protobuf equivalents."""

    def __init__(self, entity, ancestor=None, filters=(), projection=(), order=(), group_by=(),
                 limit=None, offset=0):
        """
        Initialise a new Query.

        :type entity: type
        :param entity: The entity class to use for the query. Used to derive the ``kind`` passed to datastore.

        :type ancestor: :class:`~gcloudoem.entity.Entity` or None
        :param ancestor: the ancestor to which this query's results are restricted.

        :type filters: sequence of (property_name, operator, value) tuples
        :param filters: property filters applied by this query.

        :type projection: sequence of str
        :param projection: fields to be returned as part of query results. An empty sequence means all fields.

        :type order: sequence of str
        :param order: field names used to order query results. Prepend '-' to a field name to sort it in descending
            order.

        :type group_by: sequence of str
        :param group_by: field names used to group query results.

        :type limit: int
        :param limit: number of entity results to limit this query to. None means don't limit. Defaults to None.

        :type offset: int
        :param offset: the offset into the results the first entity should be. Defaults to 0.
        """
        from gcloudoem import Entity

        if not isinstance(entity, type) or not issubclass(entity, Entity):
            raise ValueError('You must pass a valid entity class to query (one that subclasses Entity)')

        self._entity = entity
        self._ancestor = ancestor
        self._filters = list()
        self._projection = list(projection)
        self._order = list(order)
        self._group_by = list(group_by)
        self._limit = limit
        self._offset = offset
        self._has_inequality_filter = None

        for f in filters:
            self.add_filter(*f)

    @property
    def limit(self):
        return self._limit

    @limit.setter
    def limit(self, limit):
        self._limit = limit

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = offset

    def set_limits(self, offset, limit):
        """Shortcut to set the offset and the limit. Useful for slices."""
        self._offset = offset
        self._limit = limit

    def is_limited(self):
        """Has an offset or limit been applied to this query?"""
        return bool(self._offset or self._limit)

    @property
    def entity(self):
        """
        Get the Kind of the Query.

        :rtype: str
        """
        return self._entity

    @property
    def ancestor(self):
        """
        The ancestor key for the query.

        :rtype: :class:`~gcloudoem.entity.Entity` or None
        """
        return self._ancestor

    @property
    def filters(self):
        """
        Filters set on the query.

        :rtype: sequence of (property_name, operator, value) tuples.
        """
        return self._filters[:]

    def add_filter(self, property_name, operator, value):
        """
        Add a filter to the query based on a property name, operator and a value.

        Expressions take the form of::

          .add_filter('<property>', '<operator>', <value>)

        where property is the name of a property stored on the entity for this query, operator is one of ``OPERATORS``
        (ie, ``=``, ``<``, ``<=``, ``>``, ``>=``) and value is the value to filter on::

            >>> from gcloudoem import entity, properties
            >>> from gcloudoem.datastore.query import Query
            >>> class Person(entity.Entity):
            ...     name = properties.TextProperty()
            ...     age = properties.IntegerProperty()
            ...
            >>> query = Query(Person)
            >>> query.add_filter('name', '=', 'James')
            >>> query.add_filter('age', '>', 50)

        :type property_name: str
        :param property_name: A property name. Used to fetch the corresponding property off the Entity for this query.

        :type operator: str
        :param operator: One of ``=``, ``<``, ``<=``, ``>``, ``>=``. See :attr:`Query.OPERATORS`.

        :type value: int, str, bool, float, None, datetime
        :param value: The value to filter on.

        :raises: :class:`gcloudoem.exceptions.InvalidQueryError` if:
            * `operation` is not one of the specified values, or
            * query.entity doesn't have a property by the name ``property_name``, or
            * filter names `'__key__'` but passes an invalid operator (``=`` is required) or value (a Key is required).
        """
        if self.OPERATORS.get(operator) is None:
            error_message = 'Invalid expression: "%s"' % (operator,)
            choices_message = 'Please use one of: =, <, <=, >, >=.'
            raise InvalidQueryError(error_message, choices_message)

        if property_name == 'key':
            if not isinstance(value, (Key,) + six.string_types + six.integer_types):
                raise InvalidQueryError('Invalid key value "%s"' % type(value))
            if not isinstance(value, Key):
                value = Key(self._entity._meta.kind, value=value)
            if self.OPERATORS[operator] != datastore_pb.PropertyFilter.EQUAL:
                raise InvalidQueryError('Invalid operator for key: "%s"' % operator)
        elif not hasattr(self._entity, property_name):
            raise InvalidQueryError("Entity %s used in this Query doesn't have a property %s" %
                                    (self._entity._meta.kind, property_name))

        if self.OPERATORS[operator] in (
            datastore_pb.PropertyFilter.LESS_THAN_OR_EQUAL, datastore_pb.PropertyFilter.GREATER_THAN_OR_EQUAL,
        ):
            if not self._has_inequality_filter:
                self._has_inequality_filter = property_name
            elif property_name != self._has_inequality_filter:
                raise InvalidQueryError(
                    "Datastore only supports inequality operators on a single property within a query."
                )

        self._filters.append((property_name, operator, value))

    @property
    def projection(self):
        """
        Fields names returned by the query.

        :rtype: sequence of str
        :returns: Names of fields in query results.
        """
        return self._projection[:]

    @projection.setter
    def projection(self, value):
        """
        :raises: :class:`gcloudoem.exceptions.InvalidQueryError` if the property name(s) in value don't exist on the
            entity for this query.
        """
        if isinstance(value, str):
            value = [value]
        for projection in value:
            if not projection == '__key__' and not hasattr(self._entity, projection):
                raise InvalidQueryError("Entity %s used in this Query doesn't have a property %s" %
                                        (self.entity._meta.kind, projection))
        self._projection[:] = value

    def keys_only(self):
        """Set the projection to include only keys."""
        self._projection[:] = ['__key__']

    @property
    def order(self):
        """
        Names of fields used to sort query results.

        :rtype: sequence of str
        """
        return self._order[:]

    @order.setter
    def order(self, value):
        """
        Set the fields used to sort query results.

        Sort fields will be applied in the order specified.

        :type value: str or sequence of str
        :param value: Each value is a str giving the name of the property on which to sort, optionally preceded by a
            hyphen (-) to specify descending order. Omitting the hyphen implies ascending order.

        :raises: :class:`gcloudoem.exceptions.InvalidQueryError` if the property name(s) in value don't exist on the
            entity for this query.
        """
        if isinstance(value, str):
            value = [value]
        for prop_name in value:
            property = prop_name[1:] if prop_name[0] == '-' else prop_name
            if not hasattr(self._entity, property):
                raise InvalidQueryError("Entity %s used in this Query doesn't have a property %s" %
                                        (self._entity._meta.kind, prop_name))
        self._order[:] = value

    @property
    def group_by(self):
        """
        Names of fields used to group query results.

        :rtype: sequence of str
        """
        return self._group_by[:]

    @group_by.setter
    def group_by(self, value):
        """
        Set fields used to group query results.

        :type value: str or sequence of strs
        :param value: Each value is a str giving the name of a property to use to group results together.

        :raises: :class:`gcloudoem.exceptions.InvalidQueryError` if the property name(s) in value don't exist on the
            entity for this query.
        """
        if isinstance(value, str):
            value = [value]
        for prop_name in value:
            if not hasattr(self._entity, prop_name):
                raise InvalidQueryError("Entity %s used in this Query doesn't have a property %s" %
                                        (self._entity._meta.kind, prop_name))
        self._group_by[:] = value

    def __call__(self):
        """
        Execute the Query; return a :class:`Cursor` for the matching entities.

        For example::

            >>> from gcloudoem.datastore.query import Query
            >>> query = Query('Person')
            >>> query.add_filter('name', '=', 'Sally')
            >>> list(query.execute())
            [<Entity object>, <Entity object>, ...]
            >>> query.limit = 1
            >>> list(query.execute(1))
            [<Entity object>]

        For an explication of the options, see
        https://cloud.google.com/datastore/docs/concepts/queries#Datastore_Query_cursors.

        :rtype: :class:`Cursor`
        :raises: :class:`~gcloudoem.exceptions.ConnectionError` if there is no active connection.
        """
        connection = get_connection()

        return Cursor(self, connection, self.limit, self.offset)

    def clone(self):
        return self.__class__(
            self._entity, self._ancestor, self.filters, self.projection, self.order, self.group_by,
            self._limit, self._offset
        )

    def to_protobuf(self):
        """
        Convert this Query instance to the corresponding protobuf representation.

        :rtype: :class:`~gcloudoem.datastore.datastore_v1_pb2.Query`
        :returns: A protobuf query that can be sent to the protobuf API.  N.b. that it does not contain "in-flight"
            fields for ongoing query executions (cursors, offset, limit).
        """
        pb = datastore_pb.Query()

        for projection_name in self._projection:
            pb.projection.add().property.name = projection_name

        if self._entity:
            pb.kind.add().name = self._entity._meta.kind

        composite_filter = pb.filter.composite_filter
        composite_filter.operator = datastore_pb.CompositeFilter.AND

        if self.ancestor:
            ancestor_pb = utils.prepare_key_for_request(self.ancestor._properties['key'].to_protobuf(self.ancestor.key))

            # Filter on __key__ HAS_ANCESTOR == ancestor.
            ancestor_filter = composite_filter.filter.add().property_filter
            ancestor_filter.property.name = '__key__'
            ancestor_filter.operator = datastore_pb.PropertyFilter.HAS_ANCESTOR
            ancestor_filter.value.key_value.CopyFrom(ancestor_pb)

        for property_name, operator, value in self.filters:
            pb_op_enum = self.OPERATORS.get(operator)

            # Add the specific filter
            prop = self._entity._properties[property_name]
            property_filter = composite_filter.filter.add().property_filter
            property_filter.property.name = prop.db_name
            property_filter.operator = pb_op_enum

            # Set the value to filter on based on the type.
            if property_name == 'key':
                key_pb = prop.to_protobuf(value)
                property_filter.value.key_value.CopyFrom(utils.prepare_key_for_request(key_pb))
            else:
                attr, pb_value = prop.to_protobuf(value)
                utils.set_protobuf_value(property_filter.value, attr, pb_value)

        if not composite_filter.filter:
            pb.ClearField('filter')

        for prop in self.order:
            property_order = pb.order.add()

            if prop.startswith('-'):
                property_order.property.name = prop[1:]
                property_order.direction = property_order.DESCENDING
            else:
                property_order.property.name = prop
                property_order.direction = property_order.ASCENDING

        for group_by_name in self.group_by:
            pb.group_by.add().name = group_by_name

        return pb


class Cursor(object):
    """
    Represent the state of a given execution of a Query.

    This class is a generator that can be iterated.
    """

    _NOT_FINISHED = datastore_pb.QueryResultBatch.NOT_FINISHED

    _FINISHED = (
        datastore_pb.QueryResultBatch.NO_MORE_RESULTS,
        datastore_pb.QueryResultBatch.MORE_RESULTS_AFTER_LIMIT,
    )

    def __init__(self, query, connection, limit=None, offset=0, start_cursor=None, end_cursor=None):
        self._query = query
        self._connection = connection
        self._limit = limit
        self._offset = offset
        self._start_cursor = start_cursor
        self._end_cursor = end_cursor
        self._page = self._more_results = None

    def next_page(self):
        """
        Fetch a single "page" of query results.

        Low-level API for fine control: the more convenient API is to iterate on the current Iterator.

        :rtype: tuple, (entities, more_results, cursor)
        """
        pb = self._query.to_protobuf()

        start_cursor = self._start_cursor
        if start_cursor is not None:
            pb.start_cursor = base64.b64decode(start_cursor)

        end_cursor = self._end_cursor
        if end_cursor is not None:
            pb.end_cursor = base64.b64decode(end_cursor)

        if self._limit is not None:
            pb.limit = self._limit

        pb.offset = self._offset

        transaction = Transaction.current()

        query_results = self._connection.run_query(
            query_pb=pb,
            namespace=self._connection.namespace,
            transaction_id=transaction and transaction.id,
        )
        # NOTE: `query_results` contains an extra value that we don't use, namely `skipped_results`.
        #
        # NOTE: The value of `more_results` is not currently useful because the back-end always returns an enum value of
        #       MORE_RESULTS_AFTER_LIMIT even if there are no more results. See
        #       https://github.com/GoogleCloudPlatform/gcloud-python/issues/280 for discussion.
        entity_pbs, cursor_as_bytes, more_results_enum = query_results[:3]

        self._start_cursor = base64.b64encode(cursor_as_bytes)
        self._end_cursor = None

        if more_results_enum == self._NOT_FINISHED:
            self._more_results = True
        elif more_results_enum in self._FINISHED:
            self._more_results = False
        else:
            raise RuntimeError('Unexpected value returned for `more_results`.')

        self._page = [self._query.entity.from_protobuf(pb) for pb in entity_pbs]
        return self._page, self._more_results, self._start_cursor

    def __iter__(self):
        """
        Generator yielding all results matching our query.

        :rtype: sequence of :class:`gcloud.datastore.entity.Entity`
        """
        self.next_page()
        while True:
            for entity in self._page:
                yield entity
            if not self._more_results:
                break
            self.next_page()



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import warnings
import itertools
import six
import sys

from ..datastore.query import Query
from ..datastore.transaction import Transaction
from ..exceptions import GCloudError
from ..utils import VERSION_PICKLE_KEY
from .lookups import convert_lookups, LOOKUP_SEP


# The maximum number of items to display in a QuerySet.__repr__
REPR_OUTPUT_SIZE = 20


class QuerySet(object):
    """
    A set of results returned from a query.

    Internally, this just creates a :class:`~gcloudoem.datastore.query.Query`, and returns
    :class:`~gcloudoem.entity.Entity` instances from the underlying :class:`~gcloudoem.data.query.Cursor`.
    """
    def __init__(self, entity, queries=None):
        self.entity = entity
        self._result_cache = None

        self._queries = queries or []
        self._properties = None

        self._start = 0
        self._limit = None
        self._step = 1
        self._order = None
        self._is_filtered = False
        self._projection = None

    ##
    # Python data-model related functions
    ##
    def __getstate__(self):
        """Allows the QuerySet to be pickled."""
        # Force the cache to be fully populated.
        from .. import VERSION
        self._fetch_all()
        obj_dict = self.__dict__.copy()
        obj_dict[VERSION_PICKLE_KEY] = VERSION
        return obj_dict

    def __setstate__(self, state):
        from .. import VERSION
        msg = None
        pickled_version = state.get(VERSION_PICKLE_KEY)
        if pickled_version:
            current_version = VERSION
            if current_version != pickled_version:
                msg = ("Pickled queryset instance's Gcloudoem version %s does not match the current version %s." %
                       (pickled_version, current_version))
        else:
            msg = "Pickled queryset instance's Gcloudoem version is not specified."

        if msg:
            warnings.warn(msg, RuntimeWarning, stacklevel=2)

        self.__dict__.update(state)

    def __deepcopy__(self, memo):
        """Deep copy of a QuerySet doesn't populate the cache."""
        obj = self.__class__()
        for k, v in self.__dict__.items():
            if k == '_result_cache':
                obj.__dict__[k] = None
            else:
                obj.__dict__[k] = copy.deepcopy(v, memo)
        return obj

    def __getitem__(self, k):
        """Support skip and limit using getitem and slicing syntax."""
        queryset = self._clone()

        if not isinstance(k, (slice,) + six.integer_types):
            raise TypeError
        assert ((not isinstance(k, slice) and (k >= 0)) or
                (isinstance(k, slice) and (k.start is None or k.start >= 0) and
                 (k.stop is None or k.stop >= 0))), "Negative indexing is not supported."

        # Slice provided
        if isinstance(k, slice):
            qs = self._clone()
            qs._start, qs._limit, = k.start, k.stop
            if k.start and k.stop:
                qs._limit = k.stop - k.start
            qs._step = k.step
            return list(qs)[qs._start:qs._limit:qs._step]
        # Integer index provided
        elif isinstance(k, six.integer_types):
            qs = self._clone()
            return list(qs)[k]

    def __iter__(self):
        """Fills the cache by evaluating the queryset then iterates the results."""
        self._fetch_all()
        return iter(self._result_cache)

    def __nonzero__(self):
        """ Avoid to open all records in an if stmt in Py2. """
        return type(self).__bool__(self)

    def __bool__(self):
        self._fetch_all()
        return bool(self._result_cache)

    def __repr__(self):
        data = list(self[:REPR_OUTPUT_SIZE + 1])
        if len(data) > REPR_OUTPUT_SIZE:
            data[-1] = "...(remaining elements truncated)..."
        return repr(data)

    def __len__(self):
        self._fetch_all()
        return len(self._result_cache)

    ##
    # Public methods that evaluate the queryset
    ##
    def iterator(self):
        for q in self._queries:
            if self._order:
                q.order = self._order
            if self._projection:
                q.projections(self._projection)
        return itertools.chain(*[q() for q in self._queries])

    def count(self):
        """
        Returns the number of entities in the queryset as an integer.

        This results in this queryset being evaluated.
        """
        self._fetch_all()
        return len(self._result_cache)

    def get(self, *args, **kwargs):
        """Performs the query and returns a single entity matching the given keyword arguments."""
        clone = self.filter(*args, **kwargs)
        clone = clone.order_by()
        num = len(clone)
        if num == 1:
            return clone._result_cache[0]
        if not num:
            raise self.entity.DoesNotExist("%s matching query does not exist." % self.entity._meta.kind)
        raise self.entity.MultipleObjectsReturned(
            "get() returned more than one %s -- it returned %s!" % (self.entity._meta.kind, num)
        )

    def create(self, **kwargs):
        """Creates a new entity with the given kwargs, saving it to the database and returning the created entity."""
        entity = self.entity(**kwargs)
        entity.save(force_insert=True)
        return entity

    def bulk_create(self, entities):
        """
        Inserts each of the instances into the database. This does *not* call save() on each of the instances (save()
        runs in its own transaction), but the outcome is the same as if you were to call save() on each entity. Rather,
        it saves all the entities in Datastore within the one transaction.

        :type entities: iterable of :class:`~gcloudoem.entity.Entity` instances.
        :param entities: The entities to save.

        :return: The created entities
        """
        with Transaction(Transaction.SNAPSHOT) as txn:
            for entity in entities:
                txn.create(entity)

        return entities

    def get_or_create(self, defaults=None, **kwargs):
        """
        Looks up an entity with the given kwargs, creating one if necessary.

        Returns a tuple of (object, created), where created is a boolean specifying whether an object was created.

        :rtype: tuple
        """
        lookup, params = self._extract_model_params(defaults, **kwargs)
        try:
            return self.get(**lookup), False
        except self.entity.DoesNotExist:
            return self._create_object_from_params(lookup, params)

    def update_or_create(self, defaults=None, **kwargs):
        """
        Looks up an entity with the given kwargs, updating one with defaults if it exists, otherwise creates a new one.


        :rtype: tuple
        :returns: tuple (entity, created), where created is a boolean specifying whether an entity was created.
        """
        defaults = defaults or {}
        lookup, params = self._extract_model_params(defaults, **kwargs)
        try:
            entity = self.get(**lookup)
        except self.entity.DoesNotExist:
            entity, created = self._create_object_from_params(lookup, params)
            if created:
                return entity, created
        for k, v in six.iteritems(defaults):
            setattr(entity, k, v)

        entity.save()
        return entity, False

    def earliest(self, field_name=None):
        return self._earliest_or_latest(field_name=field_name, direction="")

    def latest(self, field_name=None):
        return self._earliest_or_latest(field_name=field_name, direction="-")

    def first(self):
        """
        Returns the first object of a query, returns None if no match is found.
        """
        entities = list(self)[:1]
        if entities:
            return entities[0]
        return None

    def last(self):
        """
        Returns the last object of a query, returns None if no match is found.
        """
        objects = list(self.reverse())[:1]
        if objects:
            return objects[0]
        return None

    def in_bulk(self, id_list):
        """
        Returns a dictionary mapping each of the given IDs to the entity with that ID.

        For this method, ID means the name_or_id of an entity key.
        """
        assert not self._is_limited(), "Can't 'limit' or 'offset' with in_bulk"
        if not id_list:
            return {}
        qs = self.filter(pk__in=id_list).order_by()
        return {e.key.name_or_id: e for e in qs}

    def delete(self):
        """Deletes the entities in the current QuerySet."""
        assert not self._is_limited(), "Cannot use 'limit' or 'offset' with delete."

        if self._properties is not None:
            raise TypeError("Cannot call delete() after .values() or .values_list()")

        del_query = list(self._clone())

        with Transaction(Transaction.SNAPSHOT) as txn:
            for e in iter(del_query):
                txn.delete(e)

        # Clear the result cache, in case this QuerySet gets reused.
        self._result_cache = None

    def update(self, **kwargs):
        """Updates all elements in the current QuerySet, setting all the given properties to the appropriate values."""
        assert not self._is_limited(), "Cannot update a query once a slice has been taken."
        entities = list(self)
        for e in entities:
            for name, value in kwargs.items():
                setattr(e, name, value)
        with Transaction(Transaction.SNAPSHOT) as txn:
            for e in entities:
                txn.put(e)
        self._result_cache = None
        return entities

    def exists(self):
        return self.__bool__()

    ##
    # Public methods that return an new queryset
    ##
    def all(self):
        """
        Returns a new QuerySet that is a copy of the current one. This allows a QuerySet to proxy for a model manager
        in some cases.
        """
        return self._clone()

    def filter(self, *args, **kwargs):
        """Returns a new QuerySet instance with the args ANDed to the existing set."""
        return self._filter_or_exclude(False, *args, **kwargs)

    def exclude(self, *args, **kwargs):
        """Returns a new QuerySet instance with NOT (args) ANDed to the existing set."""
        # return self._filter_or_exclude(True, *args, **kwargs)
        raise NotImplementedError

    def order_by(self, *field_names):
        """
        Returns a new QuerySet instance with the ordering changed.
        """
        assert not self._is_limited(), "Cannot reorder a query once a slice has been taken."
        obj = self._clone()
        obj._order = field_names
        return obj

    def projection(self, *properties):
        """Use a Datastore projection query. A projection query only returns certain properties as specified."""
        assert not self._is_limited(), "Cannot reorder a query once a slice has been taken."
        assert not self._projection, "projection() already used."
        clone = self._clone()
        clone._projection = properties
        return clone

    def keys_only(self):
        """Retrieve only the key property for entities in the QuerySet."""
        assert not self._is_limited(), "Cannot reorder a query once a slice has been taken."
        assert not self._projection, "Can't use keys_only() after using projection()."
        clone = self._clone()
        clone._projection = '__key__'
        return clone

    ##
    # Private methods
    ##
    def _clone(self, **kwargs):
        clone = self.__class__(entity=self.entity, queries=self._queries[:])
        clone._properties = self._properties
        clone._start = self._start
        clone._limit = self._limit
        clone._step = self._step
        clone._is_filtered = self._is_filtered
        clone._projection = self._projection

        clone.__dict__.update(kwargs)

        return clone

    def _fetch_all(self):
        """
        Evaluates this query set and populates the cache. Does nothing is the cache is already populated.

        Ultimately, this just ends up evaluating the underlying :class:`~gcloudoem.datastore.query.Quert` and saving
        the results returned by the :class:`~gcloudoem.datastore.query.Cursor`.
        """
        if not self._result_cache:
            self._result_cache = list(self.iterator())

    def _has_filters(self):
        """
        Checks if this QuerySet has any filtering going on. Note that this isn't equivalent for checking if all objects
        are present in results, for example qs[1:]._has_filters() -> False.
        """
        return self._is_filtered

    def _is_limited(self):
        """
        Checks if the QuerySet has and limiting/slicing happening.

        :return: bool
        """
        return self._start or self._limit

    def _create_object_from_params(self, lookup, params):
        """Tries to create an entity using passed params. Used by get_or_create and update_or_create."""
        try:
            return self.create(**params), True
        except GCloudError:
            exc_info = sys.exc_info()
            try:
                return self.get(**lookup), False
            except self.entity.DoesNotExist:
                pass
            six.reraise(*exc_info)

    def _extract_model_params(self, defaults, **kwargs):
        """
        Prepares `lookup` (kwargs that are valid model attributes), `params` (for creating a model instance) based on
        given kwargs; for use by get_or_create and update_or_create.

        Really, all this does is remove `lookup` args from `params` and applies `defaults` to params.
        """
        defaults = defaults or {}
        lookup = kwargs.copy()
        params = {k: v for k, v in kwargs.items() if LOOKUP_SEP not in k}
        params.update(defaults)
        return lookup, params

    def _earliest_or_latest(self, field_name=None, direction="-"):
        """Returns the latest entity, according to the model's 'get_latest_by' option or optional given field_name."""
        order_by = field_name or getattr(self.entity._meta, 'get_latest_by')
        assert bool(order_by), \
            "earliest() and latest() require either a field_name parameter or 'get_latest_by' in the model"
        assert not self._is_limited(), "Can't apply limit or slice to earliest() or latest()"
        clone = self._clone()
        clone._stop = 1
        clone.order_by('%s%s' % (direction, order_by))
        return clone.get()

    def _filter_or_exclude(self, negate, *args, **kwargs):
        if args or kwargs:
            assert not self._is_limited(), "Cannot filter a query once a slice has been taken."
        assert not negate, "Exclude not supported yet"

        clone = self._clone()
        filters = convert_lookups(**kwargs)
        for f in filters:
            # Datastore doesn't support OR queries. To do OR queries, you need to do separate queries.
            # So each AND filter gets applied to every query. Each OR filter (only the in operator) creates a new query.
            if f[1] == 'in':
                existing_filters = clone._queries[0].filters() if clone._queries else []
                clone._queries = clone._queries + [
                    Query(self.entity, filters=existing_filters + [(f[0], 'eq', value,)]) for value in f[2]
                ]
            else:
                if clone._queries:
                    for q in clone._queries:
                        q.add_filter(*f)
                else:
                    clone._queries.append(Query(self.entity, filters=[f]))
        return clone



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import, division, print_function, unicode_literals

from ..exceptions import ValidationError
from ..queryset import QuerySet


def _get_queryset(cls):
    """Inspired by django.shortcuts.*"""
    if isinstance(cls, QuerySet):
        return cls
    else:
        return cls.objects


def get_entity_or_404(cls, *args, **kwargs):
    """
    Uses get() to return an document, or raises a Http404 exception if the document does not exist.

    cls may be a Document or QuerySet object. All other passed arguments and keyword arguments are used in the get()
    query.

    Note: Like with get(), an MultipleObjectsReturned will be raised if more than one object is found.

    Inspired by django.shortcuts.*
    """
    queryset = _get_queryset(cls)
    try:
        return queryset.get(*args, **kwargs)
    except (queryset.entity.DoesNotExist, ValidationError):
        from django.http import Http404
        raise Http404('No %s matches the given query.' % queryset.entity._meta.kind)


def get_list_or_404(cls, *args, **kwargs):
    """
    Uses filter() to return a list of documents, or raise a Http404 exception if the list is empty.

    cls may be a Document or QuerySet object. All other passed arguments and keyword arguments are used in the filter()
    query.

    Inspired by django.shortcuts.*
    """
    queryset = _get_queryset(cls)
    obj_list = list(queryset.filter(*args, **kwargs))
    if not obj_list:
        from django.http import Http404
        raise Http404('No %s matches the given query.' % queryset.entity._meta.kind)
    return obj_list



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGED BY Kapiche Ltd.
# Copyright 2015 Kapiche Ltd. All rights reserved.
# Based on work by the good folk responsible for gcloud-python. Thanks folks!
# Author: Ryan Stuart<ryan@kapiche.com>
#
"""
Module to provide implicit behavior based on environment.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import socket

from six.moves.http_client import HTTPConnection

try:
    from google.appengine.api import app_identity
except ImportError:
    app_identity = None


_DATASET_ENV_VAR_NAME = 'GCLOUD_DATASET_ID'
_GCD_DATASET_ENV_VAR_NAME = 'DATASTORE_DATASET'


def compute_engine_id():
    """
    Gets the Compute Engine project ID if it can be inferred.

    Uses 169.254.169.254 for the metadata server to avoid request latency from DNS lookup.

    See https://cloud.google.com/compute/docs/metadata#metadataserver for information about this IP address. (This IP is
    also used for Amazon EC2 instances, so the metadata flavor is crucial.)

    See https://github.com/google/oauth2client/issues/93 for context about DNS latency.

    :rtype: string or ``NoneType``
    :returns: Compute Engine project ID if the metadata service is available, else ``None``.
    """
    host = '169.254.169.254'
    uri_path = '/computeMetadata/v1/project/project-id'
    headers = {'Metadata-Flavor': 'Google'}
    connection = HTTPConnection(host, timeout=0.1)

    try:
        connection.request('GET', uri_path, headers=headers)
        response = connection.getresponse()
        if response.status == 200:
            return response.read()
    except socket.error:  # socket.timeout or socket.error(64, 'Host is down')
        pass
    finally:
        connection.close()
    return None


def determine_default_dataset_id(dataset_id=None):
    """Determine default dataset ID explicitly or implicitly as fall-back.

    In implicit case, supports four environments. In order of precedence, the
    implicit environments are:

    * GCLOUD_DATASET_ID environment variable
    * DATASTORE_DATASET environment variable (for ``gcd`` testing)
    * Google App Engine application ID
    * Google Compute Engine project ID (from metadata server)

    :type dataset_id: string
    :param dataset_id: Optional. The dataset ID to use as default.

    :rtype: string or ``NoneType``
    :returns: Default dataset ID if it can be determined.
    """
    if dataset_id is None:
        dataset_id = os.getenv(_DATASET_ENV_VAR_NAME)

    if dataset_id is None:
        dataset_id = os.getenv(_GCD_DATASET_ENV_VAR_NAME)

    if dataset_id is None:
        dataset_id = compute_engine_id()

    return dataset_id



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from builtins import str as text

import six


class Key(object):
    """
    A Key for an Entity in Google Datastore.

    There should be no need to use this class directly. Rather, it should be used via
    :class:`~gcloudoem.properties.KeyProperty` on an :class:`~gcloud.entity.Entity`.

    Datastore keys have an ``id`` if they are int based, or a ``name`` if they are str based. You can use
    :prop:`name_or_id` to fetch the appropriate identifier for this key. Keys can also optionally have a parent which
    should be another :class:`Key`.

    If no id or name is specified for this key, then an int id will be auto assigned to it when the owning
    :class:`~gcloud.entity.Entity` is saved.
    """
    def __init__(self, kind, parent=None, value=None):
        """
        Initialise a new key.

        :param str kind: The Datastore king of this key.
        :param :class:`Key` parent: The parent of this key.
        :param int or st value: the name or id of this key.
        """
        self._kind = u"%s" % text(kind)  # Make SURE we have unicode
        self._parent = parent
        self._id = self._name = None
        if isinstance(value, six.string_types):
            self._name = value
        elif isinstance(value, six.integer_types):
            self._id = value

    @property
    def kind(self):
        return self._kind

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    @property
    def name_or_id(self):
        return self._name if self._name else self._id

    @property
    def parent(self):
        return self._parent

    @property
    def is_partial(self):
        return self.name_or_id is None

    @property
    def path(self):
        """
        The path of this key, oldest ancestor first.

        :rtype: list of :class:`Key`s
        :return: The full path of this :class:`Key` which includes this key itself as the last element.
        """
        path = [self]
        key = self._parent
        while key:
            path = [key] + path
            key = key.parent
        return path



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGED BY Kapiche Ltd.
# Copyright 2015 Kapiche Ltd. All rights reserved.
# Based on work by the good folk responsible for gcloud-python. Thanks folks!
# Author: Ryan Stuart<ryan@kapiche.com>
#
"""Base classes."""
from __future__ import absolute_import

import httplib2


class BaseConnection(object):
    """
    A generic connection to Google Cloud Platform.

    Subclasses should understand only the basic types in method arguments, however they should be capable of returning
    advanced types.

    A connection in datastore terms is a combination of gcloud credentials and a dataset id (project id). While its
    possible to use the same same credentials across projects in gcloud, it's rarely done and keeping all these things
    together aligns much better with a traditional database.

    If no value is passed in for ``http``, a :class:`httplib2.Http` object will be created and authorized with the
    ``credentials``. If not, the ``credentials`` and ``http`` need not be related.

    Subclasses may seek to use the private key from ``credentials`` to sign data.

    A custom (non-``httplib2``) HTTP object must have a ``request`` method which accepts the following arguments:

    * ``uri``
    * ``method``
    * ``body``
    * ``headers``

    In addition, ``redirections`` and ``connection_type`` may be used.

    Without the use of ``credentials.authorize(http)``, a custom ``http`` object will also need to be able to add a
    bearer token to API requests and handle token refresh on 401 errors.
    """
    API_BASE_URL = 'https://www.googleapis.com'
    """The base of the API call URL."""

    USER_AGENT = "gcloud-datastore-oem"
    """The user agent for requests."""

    def __init__(self, dataset, namespace, credentials=None, http=None):
        """
        :type dataset: str
        :param dataset: The gcloud Datastore dataset identifier.

        :type namespace: str
        :param namespace: The gcloud Datastore namespace to use.

        :type credentials: :class:`oauth2client.client.OAuth2Credentials` or :class:`NoneType`
        :param credentials: The OAuth2 Credentials to use for this connection.

        :type http: :class:`httplib2.Http` or class that defines ``request()``.
        :param http: An optional HTTP object to make requests.
        """
        self._dataset = dataset
        self._namespace = namespace
        self._http = http
        self._credentials = credentials

    @property
    def dataset(self):
        return self._dataset

    @property
    def namespace(self):
        return self._namespace

    @property
    def credentials(self):
        """
        Getter for current credentials.

        :rtype: :class:`oauth2client.client.OAuth2Credentials` or :class:`NoneType`
        :returns: The credentials object associated with this connection.
        """
        return self._credentials

    @property
    def http(self):
        """
        A getter for the HTTP transport used in talking to the API.

        :rtype: :class:`httplib2.Http`
        :returns: A Http object used to transport data.
        """
        if self._http is None:
            self._http = httplib2.Http()
            if self._credentials:
                self._http = self._credentials.authorize(self._http)
        return self._http



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import, division, print_function, unicode_literals


COMPARISON_OPERATORS = ('gt', 'gte', 'lt', 'lte', 'in',)
LOOKUP_SEP = '__'


def convert_lookups(**query):
    """
    Transform a query from Django-style format to Datasore format.

    :return: An iterable of filters suitable to pass to :meth:`~gcloudoem.datastore.query.Query.add_filter`.
    """
    filters = []
    for key, value in sorted(query.items()):
        parts = key.rsplit(LOOKUP_SEP)
        parts = [part for part in parts if not part.isdigit()]
        # Figure out the operator

        op = 'eq'
        if len(parts) > 1 and parts[-1] in COMPARISON_OPERATORS:
            op = parts.pop()

        # Convert to datastore notation for key. Just a nice-to-have.
        if parts[0] == 'pk':
            parts[0] = 'key'

        # Save the filter
        filters.append((parts[0], op, value,))
    return filters



# Copyright 2014 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGED BY Kapiche Ltd.
# Copyright 2015 Kapiche Ltd. All rights reserved.
# Based on work by the good folk responsible for gcloud-python. Thanks folks!
# Author: Ryan Stuart<ryan@kapiche.com>
#
"""Create / interact with a datastore transaction."""
from __future__ import absolute_import, division, print_function, unicode_literals

from . import datastore_v1_pb2 as datastore_pb, environment, utils
from .connection import get_connection
from ..properties import KeyProperty, ListProperty, ReferenceProperty

_TRANSACTIONS = utils._LocalStack()


class Transaction(object):
    """
    An abstraction representing a collected group of updates / deletes as a datastore transaction.

    Used to build up a bulk mutation.

    For example, the following snippet of code will put the two ``save`` operations and the delete operation into the
    same mutation, and send them to the server in a single API request::

      >>> from gcloudoem.datastore.transaction import Transaction
      >>> transaction = Transaction()
      >>> transaction.put(entity1)
      >>> transaction.put(entity2)
      >>> transaction.delete(entity3)
      >>> transaction.commit()

    You can also use a transaction as a context manager, in which case the ``commit`` will be called automatically if
    its block exits without raising an exception::

      >>> with Transaction() as transaction:
      ...     transaction.put(entity1)
      ...     transaction.put(entity2)
      ...     transaction.delete(key3)

    By default, no updates will be sent if the block exits with an error::

      >>> with Transaction() as transaction:
      ...   do_some_work(batch)
      ...   raise Exception() # rolls back

    After completion, you can determine if a commit succeeded or failed. For example, trying to delete a key that
    doesn't exist::

         >>> with Transaction(Transaction.SNAPSHOT) as xact:
         ...     xact.delete(entity)
         ...
         >>> xact.succeeded
         False

    If you don't want to use the context manager you can initialize a transaction manually::

      >>> transaction = Transaction()
      >>> transaction.begin()
      >>> entity = MyEntity(key=123)
      >>> transaction.put(entity)
      >>> if error:
      ...     transaction.rollback()
      ... else:
      ...     transaction.commit()

    Transactions have 3 isolation levels. They are (in order of strictness):

    #. NONE - No isolation. Everything is just wrapped in a single mutation and sent to datastore.
    #. SNAPSHOT - another transaction may not concurrently modify the data modified by this transaction.
    #. SERIALIZABLE - another transaction cannot concurrently modify the data that is read or modified by this
                      transaction.

    See https://cloud.google.com/appengine/articles/transaction_isolation for more information
    """

    NONE = None
    """No transaction isolation."""

    SNAPSHOT = False
    """Snapshot transaction isolation."""

    SERIALIZABLE = True
    """Serializable transaction isolation."""

    # Transaction states
    _INITIAL = 0
    _IN_PROGRESS = 1
    _ABORTED = 2
    _FINISHED = 3

    def __init__(self, isolation):
        """
        Construct a transaction.

        :type isolation: :class:`bool` or None. Use the class attributes as shortcuts.
        :param dataset_id: Transaction isolation level. None = NONE, False = SNAPSHOT and True = SERIALIZABLE.

        :raises: :class:`~gcloudoem.exceptions.ConnectionError` if there is no active connection.
        """
        self._connection = get_connection()
        self._mutation = datastore_pb.Mutation()
        self._auto_id_entities = []
        self._id = None
        self._status = self._INITIAL
        self._isolation = isolation

    @staticmethod
    def current():
        """Return the topmost batch / transaction, or None."""
        return _TRANSACTIONS.top

    @property
    def id(self):
        """
        Getter for the transaction ID.

        :rtype: str
        :returns: The ID of the current transaction.
        """
        return self._id

    def _add_auto_id_entity(self, entity):
        """
        Adds an entity to the list of entities to update with IDs.

        When an entity has a partial key, calling ``save()`` adds an insert_auto_id entry in the mutation.  In order to
        make sure we update the Entity once the transaction is committed, we need to keep track of which entities to
        update (and the order is important).

        When you call ``save()`` on an entity inside a transaction, if the entity has a partial key, it adds itself to
        the list of entities to be updated once the transaction is committed by calling this method.

        :type entity: :class:`~gcloudoem.entity.Entity`
        :param entity: The entity to be updated with a completed key.

        :raises: ValueError if the entity's key is already completed.
        """
        if not entity.key.is_partial:
            raise ValueError("Entity has a completed key")

        self._auto_id_entities.append(entity)

    def _assign_entity_to_mutation(self, entity, force_insert=False):
        """
        Copy ``entity`` into appropriate slot of the mutation for this transaction.

        If ``entity.key`` is incomplete, append ``entity`` to self.auto_id_entities for later fixup during ``commit``.

        :type entity: :class:`~gcloudoem.entity.Entity`
        :param entity; the entity being updated within the batch / transaction.

        :type force_insert: bool
        :param force_insert: Assign this entity to the insert mutation instead of upsert. Defaults to False.
        """
        # Prepare the key
        key = getattr(entity, 'key')
        key_pb = entity._properties['key'].to_protobuf(key)
        key_pb = utils.prepare_key_for_request(key_pb)

        # What type of mutation is this?
        if key.is_partial:  # autogen key?
            insert = self._mutation.insert_auto_id.add()
            self._auto_id_entities.append(entity)
        elif force_insert:
            insert = self._mutation.insert.add()
        else:
            insert = self._mutation.upsert.add()  # The default. Update or insert.

        # Add the key
        insert.key.CopyFrom(key_pb)

        for name, property in entity._properties.items():
            if isinstance(property, KeyProperty):  # Already dealt with the key
                continue

            value_is_reference = isinstance(property, ReferenceProperty)
            if value_is_reference:
                value = entity._data[property.name]  # We are happy with the Key, don't fetch the entity
            else:
                value = getattr(entity, name)
                if value is None:  # Nothing to save.
                    continue

            value_is_list = isinstance(property, ListProperty)
            if value_is_list and len(value) == 0:  # Nothing to save
                continue

            # Create the property
            prop = insert.property.add()
            prop.name = property.db_name
            attr, value = property.to_protobuf(value)
            utils.set_protobuf_value(prop.value, attr, value)

            if property.exclude_from_index:
                if not value_is_list:
                    prop.value.indexed = False

                for sub_value in prop.value.list_value:
                    sub_value.indexed = False

    def put(self, entity):
        """
        Store entity as part of this transaction.

        Nothing happens until :meth:`commit` is called, which happens automatically when this class is used as a context
        manager (ie using the with statement).

        .. note::
           Any existing properties for the entity will be replaced by those currently set on this instance.
           Already-stored  properties which do not correspond to keys set on this instance will be removed from the
           datastore.

        .. note::
           Property values which are "text" ('unicode' in Python2, 'str' in Python3) map to 'string_value' in the
           datastore; values which are "bytes" ('str' in Python2, 'bytes' in Python3) map to 'blob_value'.

        :type entity: :class:`~gcloud.entity.Entity`
        :param entity: the entity to be saved.
        """
        if entity.key is None:
            raise ValueError("Entity must have a key")

        self._assign_entity_to_mutation(entity)

    def create(self, entity):
        """
        Store entity as part of this transaction.

        This method is identical to :meth:`put` except that it it uses the ``insert`` slot of the Datastore mutation
        rather then ``upsert``.

        :type entity: :class:`~gcloud.entity.Entity`
        :param entity: the entity to be saved.

        :raises: ValueError if entity has no key assigned
        """
        if entity.key is None:
            raise ValueError("Entity must have a key")

        self._assign_entity_to_mutation(entity, force_insert=True)

    def delete(self, entity):
        """
        Delete entity during this transaction.

        Nothing happens until :meth:`commit` is called, which happens automatically when this class is used as a context
        manager (ie using the with statement).

        :type entity: :class:`~gcloudoem.entity.Entity`
        :param entity: the entity to be deleted.

        :raises: ValueError if key is not complete, or if the key's
                 ``dataset_id`` does not match ours.
        """
        if entity.key.is_partial:
            raise ValueError("Entity myst have a complete key")

        key_pb = utils.prepare_key_for_request(entity._properties['key'].to_protobuf(entity.key))
        self._mutation.delete.add().CopyFrom(key_pb)

    def begin(self):
        """
        Begin this transaction.

        If the isolation level is :attr:`NONE`, this is basically a no-op.

        :raises: :class:`ValueError` if the transaction has already begun.
        """
        if self._isolation == Transaction.NONE:
            return

        if self._status != self._INITIAL:
            raise ValueError('Transaction already started previously.')
        self._status = self._IN_PROGRESS
        self._id = self._connection.begin_transaction(serializable=self._isolation)

    def commit(self):
        """
        Commits the batch.

        This is called automatically upon exiting a with statement, however it can be called explicitly if you don't
        want to use a context manager.
        """
        try:
            response = self._connection.commit(self._mutation)
            # If the back-end returns without error, we are guaranteed that the response's 'insert_auto_id_key' will
            # match (length and order) the request's 'insert_auto_id` entities, which are derived from our
            # '_auto_id_entities' (no partial success).
            for new_key_pb, entity in zip(response.insert_auto_id_key, self._auto_id_entities):
                entity._data['key']._id = new_key_pb.path_element[-1].id
        finally:
            self._status = self._FINISHED
            # Clear our own ID in case this gets accidentally reused.
            self._id = None

    def rollback(self):
        """
        Rollback the transaction.

        This is a no-op for the :attr:`NONE` isolation level.

        Otherwise, this method has necessary side-effects:

        * Sets the current connection's transaction reference to None.
        * Sets the current transaction's ID to None.
        """
        try:
            if self._isolation != Transaction.NONE:
                self._connection.rollback(self._id)
        finally:
            self._status = self._ABORTED
            # Clear our own ID in case this gets accidentally reused.
            self._id = None

    def __enter__(self):
        _TRANSACTIONS.push(self)
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type is None:
                self.commit()
            else:
                self.rollback()
        finally:
            _TRANSACTIONS.pop()



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from io import open
import os
from setuptools import setup, find_packages
import sys


with open('README.rst', encoding='utf-8') as f:
    long_description = f.read()


# Use the VERSION file to get caterpillar version
with open(os.path.join(os.path.dirname(__file__), 'gcloudoem', 'VERSION')) as fh:
    version = fh.read().strip()

if sys.version_info >= (3, 0):
    install_requires = [
        "future",
        "httplib2",
        "oauth2client >= 1.4.7",
        "protobuf >= 3.0.0a1",  # Need Python 3 support
        "pycrypto",
        "pytz",
        "six",
    ]
else:
    install_requires = [
        "future",
        "httplib2",
        "oauth2client >= 1.4.7",
        "protobuf",
        "pycrypto",
        "pytz",
        "six",
    ]

setup(
    name='gcloudoem',
    version=version,
    description='gcloud-datastore-oem is a Python Object-Entity Mapper for working with Google Datastore.',
    long_description=long_description,
    url='https://github.com/Kapiche/gcloud-datastore-oem',
    author='Kapiche',
    author_email='opensource@kapiche.com',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        'Topic :: Database',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=["tests*", "docs*"]),
    install_requires=install_requires,
    extras_require={
        'dev': ['sphinx', 'sphinx_rtd_theme'],
        'test': ['nose', 'coverage', 'unittest2', 'django>=1.5.1', 'pyOpenSSL'],
        'test-py2': ['nose', 'coverage', 'unittest2', 'django>=1.5.1', 'pyOpenSSL', 'mock'],
    },
    package_data={
        'gcloudoem': ['VERSION'],
    }
)




# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import

from io import open
import os

from .datastore import credentials, connection, environment, utils, connect, query
from .entity import *
from .properties import *

version_file = open(os.path.join(os.path.dirname(__file__), 'VERSION'), encoding='utf-8')
VERSION = version_file.read().strip()
version_file.close()



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import

from ..exceptions import ValidationError


NON_FIELD_ERRORS = '__all__'


class BaseEntity(object):

    def __init__(self, **kwargs):
        """
        Initialise an entity. Values for an entity's properties can be passed as keyword or positional arguments.

        All entities store the key under a reserved ``key`` property. Property names can't start with '_'.

        :type key: :class:`Entity` or tuple or str or int
        :param key: The key for this Entity. If this is another that Entity will be assigned as the parent of this
                    Entity. If its a tuple, it should be in the form (parent_entity, key_value). Otherwise, it should be
                    the value of this key (either an int or str).
        """
        self._initialised = False
        self._data = {}  # Property values will be stored here by the property descriptors

        # Set attribute values
        for name in self._properties:
            setattr(self, name, kwargs.get(name, None))  # None causes default value to be set if present

        self._initialised = True
        self._meta.pk = self.pk  # Django compatibility hack

    @property
    def pk(self):
        return self.key.name_or_id

    def clean(self):
        """
        Hook for doing document level data cleaning before validation is run.

        Any ValidationError raised by this method will not be associated with a particular field; it will have a
        special-case association with the field defined by NON_FIELD_ERRORS.
        """
        pass

    def validate(self, clean=True):
        """
        Validate all properties on this entity and (optionally) this entity itself.

        :param bool clean: Whether to perform entity level validation via :meth:`clean()`.
        :raise :exc:`~gcloudoem.queryset.errors.ValidationError`: If validation fails.
        """
        errors = {}
        if clean:
            try:
                self.clean()
            except ValidationError as e:
                errors[NON_FIELD_ERRORS] = e
        for field, value in [(self._properties[name], self._data[name]) for name in self._properties]:
            try:
                field._validate(value)
            except ValidationError as error:
                errors[field.name] = error.errors or error
            except (ValueError, AttributeError, AssertionError) as error:
                errors[field.name] = error

        if errors:
            message = "ValidationError (%s:%s) " % (self.__class__, self.key.name_or_id)
            raise ValidationError(message, errors=errors)

    @classmethod
    def from_protobuf(cls, pb):
        """
        Get an instance of this class based on the protobug representation from the Datastore API.

        :type pb: :class:`gcloudoem.datastore.datastore_v1_pb2.Entity`
        :param pb: The protobuf representing the entity.

        :return: Instance of cls.
        """
        raise NotImplementedError



from django.conf import settings
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import UserManager
from django.db import models
try:
    from django.utils.module_loading import import_module
except ImportError:
    """Handle older versions of Django"""
    from django.utils.importlib import import_module
from django.utils.translation import ugettext_lazy as _


__all__ = (
    'get_user_document',
)


GCLOUDOEM_USER_ENTITY = getattr(
    settings, 'GCLOUDOEM_USER_DOCUMENT', 'gcloudoem.django.auth.User')


def get_user_entity():
    """
    Get the user document class used for authentication.

    This is the class defined in settings.GCLOUDOEM_USER_DOCUMENT, which defaults to `gcloudoem.django.auth.User`.
    """

    name = GCLOUDOEM_USER_ENTITY
    dot = name.rindex('.')
    module = import_module(name[:dot])
    return getattr(module, name[dot + 1:])


class GCloudUserManager(UserManager):
    """
    A User manager which allows the use of GCloudOEM documents in Django.

    To use the manager, you must tell django.contrib.auth to use GCloudUser as the user model. In you settings.py, you
    need:

        INSTALLED_APPS = (
            ...
            'django.contrib.auth',
            'gcloudoem.django.gcloud_auth',
            ...
        )
        AUTH_USER_MODEL = 'gcloud_auth.GCloudUser'

    Django will use the model object to access the custom Manager, which will replace the original queryset with
    GCloudOEM querysets.

    By default, gcloudoem.django.auth.User will be used to store users. You can specify another entity class in
    GCLOUDOEM_USER_ENTITY in your settings.py.

    The User Entity class has the same requirements as a standard custom user model:
    https://docs.djangoproject.com/en/dev/topics/auth/customizing/

    In particular, the User Entity class must define USERNAME_FIELD and REQUIRED_FIELDS.

    `AUTH_USER_MODEL` has been added in Django 1.5.
    """

    def contribute_to_class(self, model, name):
        super(GCloudUserManager, self).contribute_to_class(model, name)
        self.dj_model = self.model
        self.model = get_user_entity()

        self.dj_model.USERNAME_FIELD = self.model.USERNAME_FIELD
        username = models.CharField(_('username'), max_length=30, unique=True)
        username.contribute_to_class(self.dj_model, self.dj_model.USERNAME_FIELD)

        self.dj_model.REQUIRED_FIELDS = self.model.REQUIRED_FIELDS
        for name in self.dj_model.REQUIRED_FIELDS:
            field = models.CharField(_(name), max_length=30)
            field.contribute_to_class(self.dj_model, name)

    def get(self, *args, **kwargs):
        try:
            return self.get_query_set().get(*args, **kwargs)
        except self.model.DoesNotExist:
            # ModelBackend expects this exception
            raise self.dj_model.DoesNotExist

    @property
    def db(self):
        raise NotImplementedError

    def get_empty_query_set(self):
        return self.model.objects.none()

    def get_query_set(self):
        return self.model.objects


class GCloudUser(models.Model):
    """"
    Dummy user model for GCloud Datastore.

    GCloudUser is used to replace Django's UserManager with MongoUserManager. The actual user document class is
    gcloudoem.django.auth.User or any other document class specified in GCLOUDOEM_USER_ENTITY.

    To get the user entity class, use `get_user_entity()`.
    """
    objects = GCloudUserManager()

    class Meta:
        app_label = 'gcloud_auth'

    def set_password(self, password):
        """Doesn't do anything, but works around the issue with Django 1.6."""
        make_password(password)



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import

from .utils import camel_case_to_spaces


class Options(object):
    """
    Store the options for an entity.

    You shouldn't need to instantiate this class directly. Rather, the :class:`~gcloudoem.base.metaclasses.EntityMeta`
    will do it for you if you define a `Meta` class on your entity. For example::

        from gcloudoem import *

        class Person(Entity):
            name = TextProperty()

            class Meta:
                kind = 'special_person'

    You can access an instance on this class on any entity via the ``_meta`` attribute. The options are as follows:

    * **app_label** - The label for the application an Entity belongs to. Used for the name of the Entity in Datastore.
        Defaults to being empty.
    * **get_latest_by** - The name of an orderable property in the entity, typically a DateProperty, DateTimeProperty,
        or IntegerProperty. This specifies the default field to use in your model Manager's latest() and earliest()
        methods.
    * **indexed_properties** - Which properties should be indexed in Datastore? This should be a tuple or list of string
        property names. Defaults to all properties.
    * **kind** - What kind should be used to store this in datastore. Defaults to the Entity class name.
    * **namespace** - The namespace to use with Datastore. By default it is blank.
    * **ordering** - The default ordering for the entity, for use when obtaining lists of entities. This is a tuple or
        list of strings. Each string is a property name with an optional "-" prefix, which indicates descending order.
        Fields without a leading "-" will be ordered ascending. Use the string "?" to order randomly.
    * **queryset_class** - The class to use as the query set. Can be used to override the one set by the manager.
    * **verbose_name** - A human-readable name for the entity, singular. Defaults to a a munged version of the class
        name: CamelCase becomes camel case.
    * **verbose_name_plural** - The plural name for the entity. Defaults to verbose_name + "s".

    This class forms the Admin interface to a entity. Using it, it's possible to find out all the information about an
    entity including it's fields (:meth:`get_fields` and :meth:`get_field`), how it is stored in Datastore (including
    its kind and namespace), which fields are indexed, etc.
    """
    DEFAULT_NAMES = (
        'verbose_name', 'verbose_name_plural', 'kind', 'ordering', 'get_latest_by',
        'order_with_respect_to', 'namespace', 'indexed_properties',
    )

    def __init__(self, meta):
        # Defaults
        self.get_by_latest = None
        self.indexed_properties = []
        self.kind = ''
        self.namespace = ''
        self.ordering = []
        self.queryset_class = None
        self.verbose_name = ''
        self.verbose_name_plural = ''
        self.app_label = ''

        # Mostly internal things
        self.entity = None
        self.object_name = ''
        self.entity_name = ''
        self.meta = meta

    def contribute_to_class(self, cls, name):
        cls._meta = self
        self.entity = cls
        # First, construct the default values for these options.
        self.object_name = self.kind = cls.__name__
        self.entity_name = self.object_name.lower()
        self.verbose_name = camel_case_to_spaces(self.object_name)

        # Store the original user-defined values for each option,
        # for use when serializing the model definition
        self.original_attrs = {}

        # Next, apply any overridden values from 'class Meta'.
        if self.meta:
            meta_attrs = self.meta.__dict__.copy()
            for name in self.meta.__dict__:
                # Ignore any private attributes that Django doesn't care about.
                # NOTE: We can't modify a dictionary's contents while looping
                # over it, so we loop over the *original* dictionary instead.
                if name.startswith('_'):
                    del meta_attrs[name]
            for attr_name in self.DEFAULT_NAMES:
                if attr_name in meta_attrs:
                    setattr(self, attr_name, meta_attrs.pop(attr_name))
                    self.original_attrs[attr_name] = getattr(self, attr_name)
                elif hasattr(self.meta, attr_name):
                    setattr(self, attr_name, getattr(self.meta, attr_name))
                    self.original_attrs[attr_name] = getattr(self, attr_name)

            # verbose_name_plural is a special case because it uses a 's' by default.
            if self.verbose_name_plural is None:
                self.verbose_name_plural = self.verbose_name + 's'

            # Prepend the app_label to the kind
            if self.app_label != '':
                self.kind = "%s_%s" % (self.app_label, self.kind)

            # Any leftover attributes must be invalid.
            if meta_attrs != {}:
                raise TypeError("'class Meta' got invalid attribute(s): %s" % ','.join(meta_attrs.keys()))
        else:
            self.verbose_name_plural = self.verbose_name + 's'
        del self.meta



# Copyright (c) 2012-2015 Kapiche Ltd.
# Author: Ryan Stuart<ryan@kapiche.com>
from __future__ import absolute_import, division, print_function, unicode_literals

from functools import partial

from . import QuerySet


class QuerySetManager(object):
    """
    The default QuerySet Manager.

    Custom QuerySet Manager functions can extend this class and users can add extra queryset functionality.  Any custom
    manager methods must accept a :class:`~gcloudoem.entity.Entity` class as its first argument, and a
    :class:`~gcloudoem.queryset.QuerySet` as its second argument.

    The queryset_func function should return a :class:`~gcloudoem.queryset.QuerySet`, probably the same one that was
    passed in, but modified in some way.
    """

    get_queryset = None
    default = QuerySet

    def __init__(self, queryset_func=None):
        if queryset_func:
            self.get_queryset = queryset_func

    def __get__(self, instance, owner):
        """Descriptor for instantiating a new QuerySet object when Entity.objects is accessed."""
        if instance is not None:
            # Document class being used rather than a document object
            return self

        # owner is the document that contains the QuerySetManager
        queryset_class = owner._meta.queryset_class or self.default
        queryset = queryset_class(owner)
        if self.get_queryset:
            arg_count = self.get_queryset.func_code.co_argcount
            if arg_count == 1:
                queryset = self.get_queryset(queryset)
            elif arg_count == 2:
                queryset = self.get_queryset(owner, queryset)
            else:
                queryset = partial(self.get_queryset, owner, queryset)
        return queryset


def queryset_manager(func):
    """
    Decorator that allows you to define custom QuerySet managers on :class:`~gcloudoem.entity.Entity` classes. The
    manager must be a function that accepts a :class:`~gcloudoem.entity.Entity` class as its first argument, and a
    :class:`~gcloudoem.queryset.QuerySet` as its second argument. The method function should return a
    :class:`~gcloudoem.queryset.QuerySet`, probably the same one that was passed in, but modified in some way.
    """
    return QuerySetManager(func)
